/*
#
# Fish & Shark
# Copyright (C) 2012 Sebastian Rothe & Alexander Lambertz
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along with this program;
# if not, see <http://www.gnu.org/licenses/>.
#
*/


#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <bmpfile.h>

//Debug Output, 0 = nicht anzeigen, 1 = anzeigen
const int DEBUG_SIMPLE = 1;     //Nur fr die ntigsten Debug Outputs
const int DEBUG = 0;            //Debug Output einblenden
const int DEBUG_EXT = 0;        //Erweiterte Debug Outputs einblenden
const int DEBUG_SHOW_CHARS = 0; //Ausgabe jedes gelesenen Chars
const int DEBUG_SHOW_BMPDATA = 0; //Ausgabe des structs
const int DEBUG_SHOW_PIXEL_ERRORS = 1;      //Fr Error-Ausgaben bzgl. der Pixel

// Parameter fr Farben
rgb_pixel_t pFish = {255, 0, 0, 0};
rgb_pixel_t pShark = {0, 0, 255, 0};
rgb_pixel_t pPlankton = {0, 255, 0, 0};

const int FISH = 1;
const int SHARK = 2;
const int PLANKTON = 3;

//Error Counter
int iErrorCounter = 0;

struct bmpData{
    char posX[6];
    char posY[6];
    char val[2];
    int isPos;
    int isX;
};

rgb_pixel_t myGetColor(int value){
    if( DEBUG_EXT ) printf("COLOR should be %d.\n", value);
    if(value == FISH){
        return pFish;
    }
    if(value == SHARK){
        return pShark;
    }
    if(value == PLANKTON){
        return pPlankton;
    }
    rgb_pixel_t pError = {255, 0, 255, 0};
    if( DEBUG_SIMPLE ) iErrorCounter++;
    return pError;
}

void mySetPixel(bmpfile_t *bmp, int iValue, int dx, int dy, int iMomentaryProcess, int iSizeX, int iSizeY){
    rgb_pixel_t pColor = myGetColor(iValue);
    int x = (iSizeX * ( iMomentaryProcess % 4 ) ) + dx;
    int y = (iSizeY * ( iMomentaryProcess / 4 ) ) + dy;
    bmp_set_pixel(bmp, x, y, pColor);
    if( DEBUG_EXT ) printf("PIXEL at p(%d),l(%d|%d),g(%d|%d) set to %d.\n", iMomentaryProcess, dx, dy, x, y, iValue);
    if(( iValue == 0 || iValue > 3) && ( DEBUG_SHOW_PIXEL_ERRORS )) printf("ERROR: Pixel at p(%d),l(%d|%d),g(%d|%d) set to %d.\n", iMomentaryProcess, dx, dy, x, y, iValue);
}

int main(int argc, char **argv){

    bmpfile_t *bmp;

    // Parameter fr Anzahl der Prozesse
    int iProcess = 1;

    // Parameter fr Ozean-Gre
    int iPixelSize = 1;
    int iSizeX = 0, iSizeY = 0;

    //Parameter fr Ordner
    char *cDataDirName = "./data/";

    DIR *dirHandle;
    struct dirent **dirEntry = {0};
    int fcount = -1;
    int iFileCounter = 0;
    int i = 0;

    //Verzeichnis ffnen
    if( DEBUG_SIMPLE ) printf("DIRECTORY %s oeffnen...\n", cDataDirName);
    fcount = scandir(cDataDirName, &dirEntry, 0, alphasort);

    if(fcount < 0) {
      return 1;
    }
    if( DEBUG_SIMPLE ) printf("DIRECTORY %s geoeffnet.\n", cDataDirName);


    //Dateien einlesen
    if( DEBUG_SIMPLE ) printf("Dateien einlesen...\n");
    if( DEBUG ) printf("================================\n");
    for(i = 0; i < fcount; i++){

        //wenn es sich _nicht_ um ein Verzeichnis handelt:
        if( dirEntry[i]->d_name[0] != '.' ){
            if( DEBUG_SIMPLE ) printf("FILE %s oeffnen und einlesen...\n", dirEntry[i]->d_name);
            char cFilename[200] = "./data/";
            strcat(cFilename, dirEntry[i]->d_name);
            FILE *datei;
            char *cText;
            int iRowCounter = 0, iActualAdds = 0;

            //Datei einlesen
            datei = fopen(cFilename, "r");
            if(datei != NULL){
                if( DEBUG ) printf("FILE %s geoeffnet.\n", cFilename);
                int iFirstRow = 1;
                int iFirstValues = 0;
                int iNewRow = 0;
                char c;

                //Zeichenweise einlesen
                int iCharCounter = 0;
                //bmp_data erstellen
                struct bmpData data;
                data.isPos = 1;
                data.isX = 1;
                char cFirstRow[6];
                memset(cFirstRow, '\0', sizeof(cFirstRow)/sizeof(char));
                while( ( c = fgetc( datei ) ) != EOF ){
                    if( iFirstValues < 3 && ( ( 0 == iFileCounter ) || ( 0 == iFileCounter % iProcess ) ) ){
                        //Nur jede iProzess-te Datei darf ein neues bmp angefangen werden
                        //& die Grund-Variablen verndert werden
                        //in der ersten Zeile mssen dann spezielle Daten ausgelesen werden
                        if( ( ';' == c ) || ( ',' == c ) ){
                            //Variable kann "gefllt" werden
                            if( 0 == iFirstValues ){
                                sscanf(cFirstRow, "%d", &iProcess);
                                if( DEBUG ) printf("VARIABLE iProcess = %d\n", iProcess);
                            }
                            else if( 1 == iFirstValues ){
                                sscanf(cFirstRow, "%d", &iSizeY);
                                if( DEBUG ) printf("VARIABLE iSizeY = %d\n", iSizeY);
                            }
                            else if( 2 == iFirstValues ){
                                sscanf(cFirstRow, "%d", &iSizeX);
                                if( DEBUG ) printf("VARIABLE iSizeX = %d\n", iSizeX);
                            }
                            memset(cFirstRow, '\0', 6);
                            iFirstValues++;

                            if( 3 == iFirstValues ){
                                if( DEBUG ) printf("Neues bmp wird erstellt...\n");
                                bmp = bmp_create((iPixelSize * iSizeX * 4), (iPixelSize * iSizeY * (iProcess/4)), 24);
                                if( DEBUG ) printf("Neues bmp erstellt.\n");
                                iFirstRow = 0;
                            }
                        }
                        else{
                            strncat(cFirstRow, &c, 1);
                            if( DEBUG_EXT ) printf("first row: %s\n", cFirstRow);
                        }
                        if( DEBUG_EXT ) printf("VARIABLE iFirstValues: %d\n", iFirstValues);
                    } //if(erste Zeile vom ersten Abschnitt)
                    else if( iFirstRow ){
                        //erste Zeile darf ansonsten nicht als Pixel interpretiert werden!
                        if( ';' == c ) iFirstRow = 0;
                    }
                    else if( iRowCounter >= iSizeY ){
                        //zustzliche Informationen, nachdem alle Zeilen eingelesen wurden
                        if( ':' == c){
                            //Wert folgt
                            data.isPos = 0;
                        }
                        else if( ',' == c ){
                            data.isX = 0;
                        }
                        else if( ';' == c ){
                            //Pixel setzen
                            int iMomentaryProcess = iFileCounter % iProcess;
                            //if( DEBUG_SHOW_PIXEL_ERRORS && ( (atoi(data.val) == 0) || ( atoi(data.val) > 3 ) ) ) printf("ERROR aus Add folgend:\n");
                            mySetPixel(bmp, atoi(data.val), atoi(data.posX), atoi(data.posY), iMomentaryProcess, iSizeX, iSizeY);
                            memset(data.posX, '\0', sizeof(data.posX)/sizeof(char));
                            memset(data.posY, '\0', sizeof(data.posY)/sizeof(char));
                            memset(data.val, '\0', sizeof(data.val)/sizeof(char));
                            data.isPos = 1;
                            data.isX = 1;
                            iRowCounter++;
                            iActualAdds++;
                        }
                        else if( '\n' == c || '\0' == c ){}
                        else{
                            if( data.isPos ){
                                if( data.isX ){
                                    //X-Wert lesen
                                    strncat(data.posX, &c, 1);
                                }
                                else{
                                    //Y-Wert lesen
                                    strncat(data.posY, &c, 1);
                                }
                            }
                            else{
                                strncat(data.val, &c, 1);
                            }
                        }
                    } //if(zustzliche Zeilen nach iSizeY)
                    else{
                        //ansonsten normale Fallunterscheidung fr iSizeY Zeilen
                        if( ':' == c ){
                            //Wert folgt
                            data.isPos = 0;
                        }
                        else if( ',' == c  || ';' == c){
                            //neuer Datensatz bzw...
                            if( 0 == iNewRow){
                                data.isPos = 1;
                                int iMomentaryProcess = iFileCounter % iProcess;
                                mySetPixel(bmp, atoi(data.val), atoi(data.posX), iRowCounter, iMomentaryProcess, iSizeX, iSizeY);
                                memset(data.posX, '\0', sizeof(data.posX)/sizeof(char));
                                memset(data.val, '\0', sizeof(data.val)/sizeof(char));
                            }
                            else iNewRow = 0;
                            if( ';' == c) iRowCounter++;
                            //... neue Zeile
                        }
                        else if( '\n' == c || '\0' == c ){
                            if( '\n' == c ) iNewRow = 1;
                        }
                        else{
                            if( data.isPos ){
                                strncat(data.posX, &c, 1);
                            }
                            else{
                                strncat(data.val, &c, 1);
                            }

                            //falls ein Zeichen zuvor eine neue Zeile angefangen wurde:
                            iNewRow = 0;
                        }
                    }
                    iCharCounter++;
                    if( DEBUG_SHOW_CHARS ) printf("%d. Zeichen: %c\n", iCharCounter, c);
                    if( DEBUG_SHOW_BMPDATA ){
                        printf("bmpdata: X: %s | Y: %s | value: %s ", data.posX, data.posY, data.val);
                        if( data.isPos ) printf("| isPos");
                        if( data.isX ) printf("| isX");
                        printf("\n");
                    }
                } //while-Schleife Datei lesen
                if( DEBUG ) printf("%d Zeichen gelesen.\n", iCharCounter);
            } //if(datei != NULL)
            else{
                if( DEBUG ) printf("FILE %s konnte nicht gelesen werden.\n", cFilename);
            }
            if( DEBUG ){
                int iAdds = iRowCounter - iSizeY - 1;
                printf("%d \"Zeilen\" (%d von %d Adds) eingelesen.\n", iRowCounter, iActualAdds, iAdds);
            }
            fclose(datei);
            if( DEBUG ) printf("FILE %s geschlossen.\n", cFilename);
            if( DEBUG ) printf("-------------------------------\n");
            if( 0 == ( iFileCounter+1 ) % iProcess ){
                int iChecksum = (iFileCounter+1) % iProcess;
                if( DEBUG_EXT ) printf("CHECKSUM: 0 =?= %d, (%d+1), %d\n", iChecksum, iFileCounter, iProcess);
                if( DEBUG ) printf("OUTPUT Datei speichern...\n");
                //spezieller Fall:
                //bmp speichern und zerstren
                char cOutputName[30];
                char *cOutputFolder = "./output/";
                int iPic = ( iFileCounter + 1 ) / iProcess;

                sprintf(cOutputName, "%s%d%s", cOutputFolder, iPic, ".bmp");

                bmp_save(bmp, cOutputName);
                bmp_destroy(bmp);
                if( DEBUG_SIMPLE ) printf("OUTPUT %s gespeichert.\n", cOutputName);
                if( DEBUG_SIMPLE ) printf("================================\n\n");
            }
            //Datei-Counter erhhen
            iFileCounter++;
        }//if(ist eine Datei)
    }
    if( DEBUG_SIMPLE ) printf("================================\n");
    if( DEBUG_SIMPLE ) printf("%d Dateien eingelesen.\n", iFileCounter);
    if( DEBUG_SIMPLE ) printf("%d fehlerhafte Pixel versteckt.\n", iErrorCounter);

    //Verzeichnis schlieen
    printf("DIRECTORY %s wurde geschlossen.\n", cDataDirName);
    return 0;

}
